unit WDosEpcInOuts;

interface

uses
  SysUtils, Classes, WDosPlcUtils, WDosPlcs, WDosEpcs;

type
  TwdxEpcInOut = class (TCustomInOut)
  private
    FEpcCtrl: TwdxEpcCtrl;
    function GetBaseAddress: TEpcBaseAddress;
    function GetDemo: Boolean;
    function GetProtocol: TProtocolLength;
    function GetSlaves: TSlaves;
    function GetWatchDog: Boolean;
    function GetWatchInterval: Integer;
    procedure SetBaseAddress(Value: TEpcBaseAddress);
    procedure SetDemo(Value: Boolean);
    procedure SetProtocol(const Value: TProtocolLength);
    procedure SetSlaves(Value: TSlaves);
    procedure SetWatchDog(Value: Boolean);
    procedure SetWatchInterval(Value: Integer);
  protected
    procedure BeforeDestroy; override;
    procedure DoClose; override;
    procedure DoOpen; override;
    procedure ProcessInput(var IoData: TIoData); override;
    procedure ProcessOutput(var IoData: TIoData); override;
    procedure ResetOutput; override;
  public
    constructor Create(aOwner: TComponent); overload; override;
  published
    property BaseAddress: TEpcBaseAddress read GetBaseAddress write SetBaseAddress;
    property Demo: Boolean read GetDemo write SetDemo;
    property Protocol: TProtocolLength read GetProtocol write SetProtocol;
    property Slaves: TSlaves read GetSlaves write SetSlaves;
    property WatchDog: Boolean read GetWatchDog write SetWatchDog;
    property WatchInterval: Integer read GetWatchInterval write SetWatchInterval;
  end;
  
implementation

{
************************************** TwdxEpcInOut **************************************
}
constructor TwdxEpcInOut.Create(aOwner: TComponent);
begin
  inherited Create(aOwner);
  FEpcCtrl := TwdxEpcCtrl.Create(nil);
end;

procedure TwdxEpcInOut.BeforeDestroy;
begin
  FEpcCtrl.Free;
end;

procedure TwdxEpcInOut.DoClose;
begin
  FEpcCtrl.Close;
end;

procedure TwdxEpcInOut.DoOpen;
begin
  FEpcCtrl.Open;
end;

procedure TwdxEpcInOut.ProcessInput(var IoData: TIoData);
var
  I: Integer;
  Slave: TSlaveAddress;
  Index: Byte;
begin
  with FEpcCtrl do
  begin
  { TODO : aktuellen Zustand der Busstationen abfragen }
(*
    {Fehlermeldung der PEAS P-Baugruppe prfen}
    if Active and ErrorIUE then
      Error(ecsHardware);
*)
    { read inputs }
    for Slave :=Low(TSlaveAddress) to Succ(Slaves) do
      for Index :=0 to 6 do
      begin
        I :=Slave *8 +Index;
        Byte(IoData[I, False]) :=BusByt[Slave, Index];
      end;
  end; {with}
end;

procedure TwdxEpcInOut.ProcessOutput(var IoData: TIoData);
var
  I: Integer;
  Slave: TSlaveAddress;
  Index: Byte;
begin
  with FEpcCtrl do
  begin
    { refresh watchdog timer }
    Refresh;
    { function in run mode }
    if State =iosRun then
    begin
      { update outputs }
      for Slave :=Low(TSlaveAddress) to Succ(Slaves) do
        for Index :=0 to 5 do
        begin
          I :=Slave *8 +Index;
          BusByt[Slave, Index] :=Byte(IoData[I, True]);
        end;
    end
    else ResetOutput;
  end;
end;

procedure TwdxEpcInOut.ResetOutput;
var
  Slave: TSlaveAddress;
  Index: Byte;
begin
  { reset all outputs }
  with FEpcCtrl do
    for Slave :=Low(TSlaveAddress) to Succ(Slaves) do
      for Index :=0 to 5 do
        BusByt[Slave, Index] :=0;
end;

function TwdxEpcInOut.GetBaseAddress: TEpcBaseAddress;
begin
  Result := FEpcCtrl.BaseAddress;
end;

procedure TwdxEpcInOut.SetBaseAddress(Value: TEpcBaseAddress);
begin
  FEpcCtrl.BaseAddress := Value;
end;

function TwdxEpcInOut.GetDemo: Boolean;
begin
  Result := FEpcCtrl.Demo;
end;

procedure TwdxEpcInOut.SetDemo(Value: Boolean);
begin
  FEpcCtrl.Demo := Value;
end;

function TwdxEpcInOut.GetProtocol: TProtocolLength;
begin
  Result := FEpcCtrl.Protocol;
end;

procedure TwdxEpcInOut.SetProtocol(const Value: TProtocolLength);
begin
  FEpcCtrl.Protocol := Value;
end;

function TwdxEpcInOut.GetSlaves: TSlaves;
begin
  Result := FEpcCtrl.Slaves;
end;

procedure TwdxEpcInOut.SetSlaves(Value: TSlaves);
begin
  FEpcCtrl.Slaves := Value;
end;

function TwdxEpcInOut.GetWatchDog: Boolean;
begin
  Result := FEpcCtrl.WatchDog;
end;

procedure TwdxEpcInOut.SetWatchDog(Value: Boolean);
begin
  FEpcCtrl.WatchDog := Value;
end;

function TwdxEpcInOut.GetWatchInterval: Integer;
begin
  Result := FEpcCtrl.WatchInterval;
end;

procedure TwdxEpcInOut.SetWatchInterval(Value: Integer);
begin
  FEpcCtrl.WatchInterval := Value;
end;

end.
