unit WDosPeas;

interface

uses
  SysUtils, Classes, Forms, WDosLptCtrls;

const
  {TPort}
  LptADefault =ldLPT1;
  LptBDefault =ldLPT2;

type
  TPeasAddr =0..63;
  TPeasStatus =(psSA0, psL, psSA2, psSA3, psSA4, psSA5, psSA6);
  TStatus =set of TPeasStatus;

const
  {TPeas}
  MaxInpDefault =63;
  StatusDefault =[psSA6];

type
  TwdxPeasIO = class (TComponent)
  private
    FActive: Boolean;
    FDEV: Byte;
    FEA: Boolean;
    FIUE: Boolean;
    FLptA: TwdxLptCtrl;
    FLptB: TwdxLptCtrl;
    FOnActiveChange: TNotifyEvent;
    FSD1: Boolean;
    FSD2: Boolean;
    FStreamedActive: Boolean;
  protected
    procedure ActiveChange;
    function GetA: Byte;
    function GetDA: Byte;
    function GetDEV: Byte;
    function GetIO(Index: Integer): Byte;
    function GetIUE: Boolean;
    function GetPortA: TLptDevice;
    function GetPortB: TLptDevice;
    procedure Loaded; override;
    procedure OutControl;
    procedure SetA(Value: Byte);
    procedure SetActive(Value: Boolean);
    procedure SetDA(Value: Byte);
    procedure SetEA(Value: Boolean);
    procedure SetIO(Index: Integer; Value: Byte);
    procedure SetPortA(Value: TLptDevice);
    procedure SetPortB(Value: TLptDevice);
    procedure SetSD1(Value: Boolean);
    procedure SetSD2(Value: Boolean);
  public
    constructor Create(aOwner: TComponent); override;
    procedure Close;
    procedure Open;
    property IO[Index: Integer]: Byte read GetIO write SetIO; default;
    property A: Byte read GetA write SetA;
    property DA: Byte read GetDA write SetDA;
    property DEV: Byte read GetDEV write FDEV;
    property EA: Boolean read FEA write SetEA;
    property IUE: Boolean read GetIUE write FIUE;
    property SD1: Boolean read FSD1 write SetSD1;
    property SD2: Boolean read FSD2 write SetSD2;
  published
    property Active: Boolean read FActive write SetActive;
    property OnActiveChange: TNotifyEvent read FOnActiveChange write FOnActiveChange;
    property PortA: TLptDevice read GetPortA write SetPortA;
    property PortB: TLptDevice read GetPortB write SetPortB;
  end;
  
  TwdxPeasCtrl = class (TComponent)
  private
    FActive: Boolean;
    FAddr: Byte;
    FController: TwdxPeasIO;
    FDigit: Integer;
    FInp: Integer;
    FMaxInp: Byte;
    FMinOut: TPeasAddr;
    FOut: Integer;
    FStatus: TStatus;
  protected
    function GetAddr: TPeasAddr;
    function GetInp: Integer;
    procedure Notification(aComponent: TComponent; Operation: TOperation); override;
    procedure OutDigit;
    procedure SetActive(Value: Boolean);
    procedure SetAddr(Value: TPeasAddr);
    procedure SetController(Value: TwdxPeasIO);
    procedure SetDigit(Value: Integer);
    procedure SetOut(Value: Integer);
    procedure SetStatus(Value: TStatus);
  public
    constructor Create(aOwner: TComponent); override;
    destructor Destroy; override;
    procedure Close;
    procedure DecAddr;
    function ErrorIUE: Boolean;
    procedure IncAddr;
    procedure Open;
    procedure Refresh;
    property Addr: TPeasAddr read GetAddr write SetAddr;
    property Input: Integer read GetInp write FInp;
    property Output: Integer read FOut write SetOut;
  published
    property Active: Boolean read FActive write SetActive;
    property Controller: TwdxPeasIO read FController write SetController;
    property Digit: Integer read FDigit write SetDigit;
    property MaxInp: Byte read FMaxInp write FMaxInp default MaxInpDefault;
    property MinOut: TPeasAddr read FMinOut write FMinOut;
    property Status: TStatus read FStatus write SetStatus;
  end;
  
  EPeasError = class (Exception)
  end;
  
implementation

{
*************************************** TwdxPeasIO ***************************************
}
constructor TwdxPeasIO.Create(aOwner: TComponent);
begin
  inherited Create(aOwner);
  {LptPorts erzeugen und auf Standardwerte setzen}
  FLptA :=TwdxLptCtrl.Create(Self);
  PortA :=LptADefault;
  FLptB :=TwdxLptCtrl.Create(Self);
  PortB :=LptBDefault;
end;

procedure TwdxPeasIO.ActiveChange;
begin
  if Assigned(FOnActiveChange) then FOnActiveChange(Self);
end;

procedure TwdxPeasIO.Close;
begin
  if Active then
  begin
    {Active zurcksetzen}
    FActive :=False;
    {andere Gerte vom Schlieen informieren}
    ActiveChange;
    {beide LPT-Gerte schlieen}
    FLptA.Close;
    FLptB.Close;
  end;
end;

procedure TwdxPeasIO.Loaded;
begin
  inherited Loaded;
  try
    Active :=FStreamedActive;
  except
    {im Entwurfsnmodus Delphi die Exceptions behandeln lassen}
    if csDesigning in ComponentState then
      Application.HandleException(Self)
    else
      raise;
  end;
end;

procedure TwdxPeasIO.Open;
begin
  {beide LPT-Gerte zu ffnen}
  try
    FLptA.Open;
    FLptB.Open;
    {wenn erfolgreich, dann Active setzen}
    FActive :=True;
    {Steuersignale synchronisieren}
    OutControl;
    {andere Gerte informieren}
    ActiveChange;
  except
    {wenn Fehler beim ffnen, dann alle Gerte wieder schlieen}
    on ELptCtrlError do
    begin
      FLptB.Close;
      FLptA.Close;
      raise;
    end;
  end;
end;

procedure TwdxPeasIO.OutControl;
var
  Val: Byte;
begin
  Val :=0;
  if not FSD1 then Inc(Val, 1);
  if not FSD2 then Inc(Val, 2);
  if     FEA  then Inc(Val, 4);
  FLptA.Ctrl :=Val;
end;

function TwdxPeasIO.GetA: Byte;
begin
  Result :=FLptA.Data;
end;

procedure TwdxPeasIO.SetA(Value: Byte);
begin
  FLptA.Data :=Value;
end;

procedure TwdxPeasIO.SetActive(Value: Boolean);
begin
  if csLoading in ComponentState then
    FStreamedActive :=Value
  else
    if Value then
      Open
    else
      Close;
end;

function TwdxPeasIO.GetDA: Byte;
begin
  Result :=FLptB.Data;
end;

procedure TwdxPeasIO.SetDA(Value: Byte);
begin
  FLptB.Data :=Value;
end;

function TwdxPeasIO.GetDEV: Byte;
begin
  Result :=((FLptB.Status and $F0) or (FLptA.Status shr 4)) xor $77;
end;

procedure TwdxPeasIO.SetEA(Value: Boolean);
begin
  if FEA <> Value then
  begin
    FEA := Value;
    OutControl;
  end;
end;

function TwdxPeasIO.GetIO(Index: Integer): Byte;
begin
  {Schreibsignal lschen (zur Sicherheit)}
  EA :=False;
  {Portadresse anwhlen}
  A :=Index;
  {Port-Daten lesen}
  Result :=DEV;
end;

procedure TwdxPeasIO.SetIO(Index: Integer; Value: Byte);
begin
  {Schreibsignal lschen (zur Sicherheit) }
  EA :=False;
  {Portadresse anwhlen}
  A :=Index;
  {Sendedaten setzen}
  DA :=Value;
  {Schreibimpuls erzeugen}
  EA :=True;
  EA :=False;
end;

function TwdxPeasIO.GetIUE: Boolean;
begin
  Result :=(FLptA.Status and $08) >0;
end;

function TwdxPeasIO.GetPortA: TLptDevice;
begin
  Result :=FLptA.Device;
end;

procedure TwdxPeasIO.SetPortA(Value: TLptDevice);
begin
  FLptA.Device :=Value;
end;

function TwdxPeasIO.GetPortB: TLptDevice;
begin
  Result :=FLptB.Device;
end;

procedure TwdxPeasIO.SetPortB(Value: TLptDevice);
begin
  FLptB.Device :=Value;
end;

procedure TwdxPeasIO.SetSD1(Value: Boolean);
begin
  if FSD1 <> Value then
  begin
    FSD1 := Value;
    OutControl;
  end;
end;

procedure TwdxPeasIO.SetSD2(Value: Boolean);
begin
  if FSD2 <> Value then
  begin
    FSD2 := Value;
    OutControl;
  end;
end;

{
************************************** TwdxPeasCtrl **************************************
}
constructor TwdxPeasCtrl.Create(aOwner: TComponent);
begin
  inherited Create(aOwner);
  {setzen der Default-Werte}
  FMaxInp :=MaxInpDefault;
  FStatus :=StatusDefault;
end;

destructor TwdxPeasCtrl.Destroy;
begin
  Close;
  inherited Destroy;
end;

procedure TwdxPeasCtrl.Close;
begin
  FActive :=False;
  if Assigned(FController) then
    FController.Close;
end;

procedure TwdxPeasCtrl.DecAddr;
begin
  Dec(FAddr);
  if (FAddr and $0F) =$0F then Dec(FAddr, 6);
end;

function TwdxPeasCtrl.ErrorIUE: Boolean;
begin
  Result :=FActive and not FController.IUE;
end;

procedure TwdxPeasCtrl.IncAddr;
begin
  Inc(FAddr);
  if (FAddr and $0F) =$0A then Inc(FAddr, 6);
end;

procedure TwdxPeasCtrl.Notification(aComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(aComponent, Operation);
  if (Operation =opRemove) and (aComponent =FController) then
  begin
    Close;
    FController :=nil;
  end;
end;

procedure TwdxPeasCtrl.Open;
begin
  if Assigned(FController) then
  begin
    FController.Open;
    FActive :=True;
    OutDigit;
  end
  else
    raise EPeasError.Create('kein Controller zugeordnet');
end;

procedure TwdxPeasCtrl.OutDigit;
var
  AStatus: TStatus;
begin
  {7-Segment-Anzeige auf A63 setzen}
  if FActive then
    {Keine Ausgnge im Fehler-Modus setzen,
     damit Fehleranzeige an der PEAS nicht
     versehentlich gelscht wird. }
    if not ErrorIUE then
    begin
      {altes Statusregister merken}
      AStatus :=Status;
      {Fehlerberwachung abschalten}
      Status :=Status +[psSA5];
      {Anzeigewert ausgeben}
      FController[$63] :=FDigit;
      {Statusregister auf alten Wert zurck}
      Status :=AStatus;
    end;
end;

procedure TwdxPeasCtrl.Refresh;
begin
  {Refresh-Port (Addr $7F) ansprechen}
  if FActive then
    FController[$7F] :=0;
end;

procedure TwdxPeasCtrl.SetActive(Value: Boolean);
begin
  if Value then
    Open
  else
    Close;
end;

function TwdxPeasCtrl.GetAddr: TPeasAddr;
begin
  {Umwandlung der internen Hex-Form in Dez.}
  Result :=(FAddr div 16) *10 +(FAddr mod 16);
end;

procedure TwdxPeasCtrl.SetAddr(Value: TPeasAddr);
var
  ValHex: Byte;
begin
  {Umwandlung von Value in interne Hex-Form}
  ValHex :=(Value div 10) *16 +(Value mod 10);
  if ValHex <>FAddr then
    FAddr :=ValHex;
end;

procedure TwdxPeasCtrl.SetController(Value: TwdxPeasIO);
begin
  if FController <> Value then
  begin
    if not FActive then
    begin
      FController :=Value;
      if Assigned(FController) then
        FController.FreeNotification(Self);
    end
    else
      raise EPeasError.Create(
          'Operation bei aktivem Controller nicht ausfhrbar');
  end;
end;

procedure TwdxPeasCtrl.SetDigit(Value: Integer);
begin
  {Ansteuerung der 7-Segment-Anzeige auf A63}
  Value :=Value and $0F;
  if FDigit <>Value then
  begin
    FDigit := Value;
    OutDigit;
  end;
end;

function TwdxPeasCtrl.GetInp: Integer;
begin
  {Daten von Adresse FAddr lesen}
  if FActive then
    Result :=FController[FAddr]
  else
    Result :=0;
end;

procedure TwdxPeasCtrl.SetOut(Value: Integer);
begin
  {Daten auf Port FAddr ausgeben}
  FOut := Value;
  if FActive then
    FController[FAddr] :=Value;
end;

procedure TwdxPeasCtrl.SetStatus(Value: TStatus);
begin
  {Ausgabe der Status-Signale auf A62}
  if FStatus <>Value then
  begin
  FStatus := Value;
    {Statusregister setzen}
    if FActive then
      FController[$62] :=Byte(FStatus);
  end;
end;

end.
