
{*******************************************************}
{                                                       }
{       Borland Delphi Run-time Library                 }
{       WDOSX WFSE file Interface Unit                  }
{                                                       }
{       for Delphi 4, 5 WDosX DOS-Extender              }
{       Copyright (c) 2000 by Immo Wache                }
{       e-mail: immo.wache@t-online.de                  }
{                                                       }
{*******************************************************}

{
  Version history:
    IWA 11/20/00 Version 1.1
      new behaviour
      uses WDosCheck to check first for WDosX presence
}

unit WDosWfseStreams;

interface

uses
  WDosCheck, WDosWindows, WDosSysUtils, WDosClasses;

type

{ TWfseStream class }

  TWfseStream = class (TStream)
  private
    FHandle: Integer;
  public
    constructor Create(const FileName: string);
    destructor Destroy; override;
    function Read(var Buffer; Count: Longint): LongInt; override;
    function Seek(Offset: Longint; Origin: Word): LongInt; override;
    function Write(const Buffer; Count: Longint): LongInt; override;
    property Handle: Integer read FHandle;
  end;

function WfseFileSeek(Handle, AOffset, Origin: Integer): Integer;

var
  WfseVersion: Integer =0;

implementation

uses
  WDosConsts, WDosResStrings;

var
  DataSelector: LongInt;

function WfseFileOpen(FileName: PChar): Integer; assembler;
asm
        MOV     EDX,EAX
        MOV     EAX,3D00FFFDH
        INT     21H
        JNC     @@0
        SBB     EAX,EAX
  @@0:
end;

procedure WfseFileClose(FHandle: Integer); assembler;
asm
        PUSH    EBX
        MOV     EBX,EAX
        MOV     EAX,3E00FFFDH
        INT     21H
        POP     EBX
end;

function WfseFileRead(Handle: Integer; var Buffer; Count: LongWord): Integer;
assembler;
asm
        PUSH    EBX
        MOV     EBX,EAX
        MOV     EAX,3F00FFFDH
        INT     21H
        JNC     @@0
        SBB     EAX,EAX
  @@0:
        POP     EBX
end;

function WfseFileSeek(Handle, AOffset, Origin: Integer): Integer; assembler;
asm
        PUSH    EBX
        MOV     EBX,EAX
        MOV     EAX,ECX
        AND     EAX,03H
        CMP     EAX,03H
        JNZ     @@0
        XOR     EAX,EAX
  @@0:
        SHL     EAX,10H
        ADD     EAX,4200FFFDH
        MOV     ECX,EDX
        SHR     ECX,10H
        INT     21H
        JNC     @@1
        SBB     EAX,EAX
        JNE     @@2
  @@1:
        SHL     EDX,10H
        AND     EAX,0FFFFH
        ADD     EAX,EDX
  @@2:
        POP     EBX
end;

{
********************************* TWfseStream **********************************
}
constructor TWfseStream.Create(const FileName: string);
begin
  FHandle := WfseFileOpen(PChar(FileName));
  if FHandle <0 then
    raise EFOpenError.CreateResFmt(@SFOpenError, [FileName]);
end;

destructor TWfseStream.Destroy;
begin
  if FHandle >=0 then WfseFileClose(FHandle);
  inherited Destroy;
end;

function TWfseStream.Read(var Buffer; Count: Longint): LongInt;
begin
  Result := WfseFileRead(FHandle, Buffer, Count);
  if Result <0 then Result := 0;
end;

function TWfseStream.Seek(Offset: Longint; Origin: Word): LongInt;
begin
  Result := WfseFileSeek(FHandle, Offset, Origin);
end;

function TWfseStream.Write(const Buffer; Count: Longint): LongInt;
begin
  raise EStreamError.CreateRes(@SCantWriteWfseStreamError);
  Result :=0;
end;

var
  Ident: LongWord;
  Carry: Integer;

initialization
  asm
        PUSH    EBX
        MOV     EAX,0000FFFDH
        INT     21H
        MOV     Ident,EAX
        MOV     WfseVersion,EBX
        SBB     EAX,EAX
        MOV     Carry,EAX
        POP     EBX
        MOV	DataSelector,DS
  end;
  if (Ident <>$57465345) { 'WFSE' } or (Carry <>0) then
  begin
    Writeln(SCantInitializeWfse);
    Halt(1);
  end;

end.
